package uah.regex.gp.breed;

import uah.regex.Regex;
import uah.regex.gp.nodes.RegexNode;
import uah.regex.gp.problem.GPRegexProblem;
import uah.regex.problem.RegexProblem;
import ec.BreedingPipeline;
import ec.EvolutionState;
import ec.Individual;
import ec.gp.GPBreedingPipeline;
import ec.gp.GPFunctionSet;
import ec.gp.GPIndividual;
import ec.gp.GPInitializer;
import ec.gp.GPNode;
import ec.gp.GPNodeBuilder;
import ec.gp.GPNodeSelector;
import ec.gp.GPTree;
import ec.gp.GPType;
import ec.gp.koza.GPKozaDefaults;
import ec.gp.koza.MutationPipeline;
import ec.util.Parameter;

@SuppressWarnings("serial")
public class RegexMutationPipeline extends GPBreedingPipeline {
	public static final String P_NUM_TRIES = "tries";
	public static final String P_NODE_TRIES = "node-tries";
	public static final String P_MAXDEPTH = "maxdepth";
	public static final String P_MUTATION = "mutate";
	public static final String P_BUILDER = "build";
	public static final String P_EQUALSIZE = "equal";
	public static final String P_KOZA = "koza";
	public static final String P_POINT = "point";
	public static final int INDS_PRODUCED = 1;
	public static final int NUM_SOURCES = 1;

	boolean verbose = false;

	/** How the pipeline chooses a subtree to mutate */
	public GPNodeSelector nodeselect;

	/** How the pipeline builds a new subtree */
	public GPNodeBuilder builder;

	/** The number of times the pipeline tries to build a valid mutated
	    tree before it gives up and just passes on the original */
	int numTries;

	int nodeTries;

	/** The maximum depth of a mutated tree */
	int maxDepth;

	/** Do we try to replace the subtree with another of the same size? */
	boolean equalSize;
	
	boolean koza;
	
	boolean point;

	/** Is our tree fixed?  If not, this is -1 */
	int tree;

	public Parameter defaultBase() {
		return GPKozaDefaults.base().push(P_MUTATION);
	}

	public int numSources() {
		return NUM_SOURCES;
	}

	public Object clone() {
		RegexMutationPipeline c = (RegexMutationPipeline) (super.clone());

		// deep-cloned stuff
		c.nodeselect = (GPNodeSelector) (nodeselect.clone());

		return c;
	}

	public void setup(final EvolutionState state, final Parameter base) {
		super.setup(state, base);

		Parameter def = defaultBase();
		Parameter p = base.push(P_NODESELECTOR).push("" + 0);
		Parameter d = def.push(P_NODESELECTOR).push("" + 0);

		nodeselect = (GPNodeSelector) (state.parameters
				.getInstanceForParameter(p, d, GPNodeSelector.class));
		nodeselect.setup(state, p);

		p = base.push(P_BUILDER).push("" + 0);
		d = def.push(P_BUILDER).push("" + 0);

		builder = (GPNodeBuilder) (state.parameters.getInstanceForParameter(p,
				d, GPNodeBuilder.class));
		builder.setup(state, p);

		numTries = state.parameters.getInt(base.push(P_NUM_TRIES), def
				.push(P_NUM_TRIES), 1);
		if (numTries == 0)
			state.output
			.fatal("Mutation Pipeline has an invalid number of tries (it must be >= 1).",
					base.push(P_NUM_TRIES), def.push(P_NUM_TRIES));

		nodeTries = state.parameters.getInt(base.push(P_NODE_TRIES), def
				.push(P_NODE_TRIES), 0);
		if (nodeTries == 0)
			state.output
			.fatal("Mutation Pipeline has an invalid number of node tries (it must be >= 1).",
					base.push(P_NODE_TRIES), def.push(P_NODE_TRIES));

		maxDepth = state.parameters.getInt(base.push(P_MAXDEPTH), def
				.push(P_MAXDEPTH), 1);
		if (maxDepth == 0)
			state.output.fatal("The Mutation Pipeline " + base
					+ "has an invalid maximum depth (it must be >= 1).", base
					.push(P_MAXDEPTH), def.push(P_MAXDEPTH));

		equalSize = state.parameters.getBoolean(base.push(P_EQUALSIZE), def
				.push(P_EQUALSIZE), false);

		tree = TREE_UNFIXED;
		if (state.parameters.exists(base.push(P_TREE).push("" + 0), def.push(
				P_TREE).push("" + 0))) {
			tree = state.parameters.getInt(base.push(P_TREE).push("" + 0), def
					.push(P_TREE).push("" + 0), 0);
			if (tree == -1)
				state.output
				.fatal("Tree fixed value, if defined, must be >= 0");
		}
		
		koza = state.parameters.getBoolean(base.push(P_KOZA), def
				.push(P_KOZA), true);
		point = state.parameters.getBoolean(base.push(P_POINT), def
				.push(P_KOZA), true);
	}

	/** Returns true if inner1 can feasibly be swapped into inner2's position */
	private boolean verifyPoints(GPNode inner1, GPNode inner2) {
		// We know they're swap-compatible since we generated inner1
		// to be exactly that.  So don't bother.

		// next check to see if inner1 can fit in inner2's spot
		if (inner1.depth() + inner2.atDepth() > maxDepth)
			return false;

		// checks done!
		return true;
	}

	public int produce(final int min, final int max, final int start,
			final int subpopulation, final Individual[] inds,
			final EvolutionState state, final int thread) {
		RegexProblem problem = ((GPRegexProblem) state.evaluator.p_problem).problem;
		
		// grab individuals from our source and stick 'em right into inds.
		// we'll modify them from there
		int n = sources[0].produce(min, max, start, subpopulation, inds, state,
				thread);

		// now let's mutate 'em
		for (int q = start; q < start + n; q++) {
			GPIndividual i = (GPIndividual) inds[q];

			if (tree != TREE_UNFIXED && (tree < 0 || tree >= i.trees.length))
				// uh oh
				state.output
				.fatal("GP Mutation Pipeline attempted to fix tree.0 to a value which was out of bounds of the array of the individual's trees.  Check the pipeline's fixed tree values -- they may be negative or greater than the number of trees in an individual");

			int t;
			// pick random tree
			if (tree == TREE_UNFIXED)
				if (i.trees.length > 1)
					t = state.random[thread].nextInt(i.trees.length);
				else
					t = 0;
			else
				t = tree;

			// prepare the nodeselector
			nodeselect.reset();

			// pick a node

			GPNode p1 = null; 

			String regex = null;
			float fitness = 0f;

			for (int y = 0; y < nodeTries; y++ ) {
				// pick a node in individual 1
				p1 = nodeselect.pickNode(state, subpopulation, thread, i, i.trees[t]);
				regex = ((Regex) p1).getRegex();
				fitness = ((GPRegexProblem) state.evaluator.p_problem).problem.fitness(regex);
				if (verbose) {
					System.out.printf("Selecting node (%d): %s : %f\n", y, regex, fitness);
					System.out.flush();
				}

				if (fitness == 0) break;
			}

			if (fitness != 0.0) return n; // We do not perform mutation

			// Select mutation operator
			// Children nodes score more than 0? If so, mutate all the branch
			boolean useKoza = true;
			boolean usePoint = true;
			if (verbose) System.out.print("\tChildren --> ");
			
			for (GPNode node : p1.children) {
				String childRegex = ((Regex) node).getRegex();
				fitness = problem.fitness(childRegex);
				if (verbose) System.out.printf(" %s = %.2f", childRegex, fitness);
				if (problem.fitness(childRegex) != 0f) useKoza = false;
				if (problem.fitness(childRegex) == 0f) usePoint = false;
			}
			
			if (verbose) System.out.println();
			
			if (useKoza && koza) { 
				if (verbose) System.out.println("\tMutating KOZA");
				mutateKoza(min, max, start, subpopulation, inds, state, thread, p1, t, i, q);
				return n;
			}
			
			if (usePoint && point) { 
				if (verbose) System.out.println("\tMutating POINT");
				mutateOne(min, max, start, subpopulation, inds, state, thread, p1, t, i, q);
				return n;
			}
		}
		return n;
	}

	private void mutateKoza(final int min, final int max, final int start,
			final int subpopulation, final Individual[] inds,
			final EvolutionState state, final int thread, GPNode p1,
			int t, GPIndividual i, int q) {
		GPNode p2 = null;
		boolean res = false;
		GPInitializer initializer = ((GPInitializer) state.initializer);

		for (int x = 0; x < numTries; x++) {
			// System.out.println("\tMutating");
			// generate a tree swap-compatible with p1's position
			int size = GPNodeBuilder.NOSIZEGIVEN;
			if (equalSize)
				size = p1.numNodes(GPNode.NODESEARCH_ALL);

			p2 = builder.newRootedTree(state, p1.parentType(initializer),
					thread, p1.parent,
					i.trees[t].constraints(initializer).functionset,
					p1.argposition, size);

			// check for depth and swap-compatibility limits
			res = verifyPoints(p2, p1); // p2 can fit in p1's spot -- the order
			// is important!

			// did we get something that had both nodes verified?
			if (res)
				break;
		}

		GPIndividual j;

		if (sources[0] instanceof BreedingPipeline)
			// it's already a copy, so just smash the tree in
		{
			j = i;
			if (res) // we're in business
			{
				p2.parent = p1.parent;
				p2.argposition = p1.argposition;
				if (p2.parent instanceof GPNode)
					((GPNode) (p2.parent)).children[p2.argposition] = p2;
				else
					((GPTree) (p2.parent)).child = p2;
				j.evaluated = false; // we've modified it
			}
		} else // need to clone the individual
		{
			j = (GPIndividual) (i.lightClone());

			// Fill in various tree information that didn't get filled in there
			j.trees = new GPTree[i.trees.length];

			// at this point, p1 or p2, or both, may be null.
			// If not, swap one in. Else just copy the parent.
			for (int x = 0; x < j.trees.length; x++) {
				if (x == t && res) // we've got a tree with a kicking cross
					// position!
				{
					j.trees[x] = (GPTree) (i.trees[x].lightClone());
					j.trees[x].owner = j;
					j.trees[x].child = i.trees[x].child
					.cloneReplacingNoSubclone(p2, p1);
					j.trees[x].child.parent = j.trees[x];
					j.trees[x].child.argposition = 0;
					j.evaluated = false;
				} // it's changed
				else {
					j.trees[x] = (GPTree) (i.trees[x].lightClone());
					j.trees[x].owner = j;
					j.trees[x].child = i.trees[x].child.cloneReplacing();
					j.trees[x].child.parent = j.trees[x];
					j.trees[x].child.argposition = 0;
				}
			}
		}

		if (verbose) {
			System.out.printf("\tMutated node: %s\n", ((Regex) p2).getRegex());
			System.out.flush();
		}
		// add the new individual, replacing its previous source
		inds[q] = j;
	}
	

    public int mutateOne(final int min, final int max, final int start,
			final int subpopulation, final Individual[] inds,
			final EvolutionState state, final int thread, GPNode p1, int t,
			GPIndividual i, int q) {
		// grab n individuals from our source and stick 'em right into inds.
		// we'll modify them from there
		int n = sources[0].produce(min, max, start, subpopulation, inds, state,
				thread);
		GPInitializer initializer = ((GPInitializer) state.initializer);

		if (tree != TREE_UNFIXED && (tree < 0 || tree >= i.trees.length))
			// uh oh
			state.output
					.fatal("MutateOneNodePipeline attempted to fix tree.0 to a value which was out of bounds of the array of the individual's trees.  Check the pipeline's fixed tree values -- they may be negative or greater than the number of trees in an individual");

		// prepare the nodeselector
		nodeselect.reset();

		GPNode p2 = null;

		// generate a tree with a new root but the same children,
		// which we will replace p1 with
		GPType type;
		type = p1.parentType(initializer);

		p2 = (GPNode) (pickCompatibleNode(p1, i.trees[t]
				.constraints(initializer).functionset, state, type, thread))
				.lightClone();

		// if it's an ERC, let it set itself up
		p2.resetNode(state, thread);
		
		// p2's parent and argposition will be set automatically below
		GPIndividual j;

		if (sources[0] instanceof BreedingPipeline)
		// it's already a copy, so just smash the tree in
		{
			j = i;
			p1.replaceWith(p2);
			j.evaluated = false;
		} else {
			j = (GPIndividual) (i.lightClone());

			// Fill in various tree information that didn't get filled in there
			j.trees = new GPTree[i.trees.length];

			for (int x = 0; x < j.trees.length; x++) {
				if (x == t) // we've got a tree with a kicking cross position!
				{
					j.trees[x] = (GPTree) (i.trees[x].lightClone());
					j.trees[x].owner = j;
					j.trees[x].child = i.trees[x].child.cloneReplacingAtomic(
							p2, p1);
					j.trees[x].child.parent = j.trees[x];
					j.trees[x].child.argposition = 0;
					j.evaluated = false;
				} // it's changed
				else {
					j.trees[x] = (GPTree) (i.trees[x].lightClone());
					j.trees[x].owner = j;
					j.trees[x].child = i.trees[x].child.cloneReplacing();
					j.trees[x].child.parent = j.trees[x];
					j.trees[x].child.argposition = 0;
				}
			}
		}

		
		if (p2 == null) {
			System.out.println("Warning: null p2 node in RegexMutationPipeline::mutationOne()");
		}
		
		if (verbose) {
			System.out.printf("\tMutated node: %s\n", ((Regex) p2).getRegex());
			System.out.flush();
		}
		
		// add the new individual, replacing its previous source
		inds[q] = j;
		return n;
	}

	/**
	 * Returns a node which is swap-compatible with returntype, and whose
	 * arguments are swap-compatible with the current children of original. You
	 * need to clone this node.
	 */
	private GPNode pickCompatibleNode(final GPNode original,
			final GPFunctionSet set, final EvolutionState state,
			final GPType returntype, final int thread) {
		// an expensive procedure: we will linearly search for a valid node
		int numValidNodes = 0;

		int type = returntype.type;
		GPInitializer initializer = ((GPInitializer) state.initializer);
		int len = original.constraints(initializer).childtypes.length;
		boolean failed;

		if (initializer.numAtomicTypes + initializer.numSetTypes == 1) // easy
			numValidNodes = set.nodesByArity[type][len].length;
		else
			for (int x = 0; x < set.nodesByArity[type][len].length; x++) // ugh, the hard way -- nodes swap-compatible with type, and of arity len
			{
				failed = false;
				for (int y = 0; y < set.nodesByArity[type][len][x]
						.constraints(initializer).childtypes.length; y++)
					if (!set.nodesByArity[type][len][x]
							.constraints(initializer).childtypes[y]
							.compatibleWith(initializer, original.children[y]
									.constraints(initializer).returntype)) {
						failed = true;
						break;
					}
				if (!failed)
					numValidNodes++;
			}

		// we must have at least success -- the node itself.  Otherwise we're
		// in deep doo-doo.

		// now pick a random node number
		int nodenum = state.random[thread].nextInt(numValidNodes);

		// find and return that node
		int prosnode = 0;

		if (numValidNodes == set.nodesByArity[type][len].length) // easy
			return set.nodesByArity[type][len][nodenum];
		else
			for (int x = 0; x < set.nodesByArity[type][len].length; x++) // ugh, the hard way -- nodes swap-compatible with type, and of arity len
			{
				failed = false;
				for (int y = 0; y < set.nodesByArity[type][len][x]
						.constraints(initializer).childtypes.length; y++)
					if (!set.nodesByArity[type][len][x]
							.constraints(initializer).childtypes[y]
							.compatibleWith(initializer, original.children[y]
									.constraints(initializer).returntype)) {
						failed = true;
						break;
					}
				if (!failed) {
					if (prosnode == nodenum) // got it!
						return set.nodesByArity[type][len][x];
					prosnode++;
				}
			}

		// should never be able to get here
		throw new InternalError(); // whoops!
	}
}
