package uah.regex.weka;

import weka.core.Instances;
import weka.core.converters.ConverterUtils.DataSource;
import weka.core.Utils;
import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;

import java.io.FileWriter;
import java.util.Calendar;
import java.util.Random;

/**
 * Performs multiple runs of cross-validation.
 *
 * Command-line parameters:
 * <ul>
 *    <li>-t filename - the dataset to use</li>
 *    <li>-x int - the number of folds to use</li>
 *    <li>-r int - the number of runs to perform</li>
 *    <li>-c int - the class index, "first" and "last" are accepted as well;
 *    "last" is used by default</li>
 *    <li>-W classifier - classname and options, enclosed by double quotes; 
 *    the classifier to cross-validate</li>
 *    <li>-o <output> - Prints the output statistics into <output>
 * </ul>
 *
 * Example command-line:
 * <pre>
 * java CrossValidationMultipleRuns -t labor.arff -c last -x 10 -r 10 -W "weka.classifiers.trees.J48 -C 0.25"
 * </pre>
 *
 * @author FracPete (fracpete at waikato dot ac dot nz)
 */
public class EvaluateECJ {

  /**
   * Performs the cross-validation. See Javadoc of class for information
   * on command-line parameters.
   *
   * @param args        the command-line parameters
   * @throws Excecption if something goes wrong
   */
  public static void main(String[] args) throws Exception {
    // loads data and set class index
    Instances data = DataSource.read(Utils.getOption("t", args));
    String output = Utils.getOption("o", args);
    String clsIndex = Utils.getOption("c", args);
    if (clsIndex.length() == 0)
      clsIndex = "last";
    if (clsIndex.equals("first"))
      data.setClassIndex(0);
    else if (clsIndex.equals("last"))
      data.setClassIndex(data.numAttributes() - 1);
    else
      data.setClassIndex(Integer.parseInt(clsIndex) - 1);

    // classifier
    String[] tmpOptions;
    String classname;
    tmpOptions     = Utils.splitOptions(Utils.getOption("W", args));
    classname      = tmpOptions[0];
    tmpOptions[0]  = "";
    Classifier cls = (Classifier) Utils.forName(Classifier.class, classname, tmpOptions);

    // other options
    int runs  = Integer.parseInt(Utils.getOption("r", args));
    int folds = Integer.parseInt(Utils.getOption("x", args));

    System.out.print("Weka simulation ");
    
    FileWriter writer = new FileWriter(output, true);
    writer.write("=== Setup run ===\n");
    writer.write("Classifier: " + cls.getClass().getName() + " " + Utils.joinOptions(cls.getOptions()));
    writer.write("\nDataset: " + data.relationName());
    writer.write("\nFolds: " + folds);
    writer.write("\n=================\n\n");
    // perform cross-validation
    for (int i = 0; i < runs; i++) {
      // randomize data
      Calendar cal = Calendar.getInstance();

      int seed = (int) cal.getTimeInMillis();
      Random rand = new Random(seed);
      Instances randData = new Instances(data);
      randData.randomize(rand);
      if (randData.classAttribute().isNominal())
        randData.stratify(folds);

      Evaluation eval = new Evaluation(randData);
      for (int n = 0; n < folds; n++) {
        Instances train = randData.trainCV(folds, n);
        Instances test = randData.testCV(folds, n);
        // the above code is used by the StratifiedRemoveFolds filter, the
        // code below by the Explorer/Experimenter:
        // Instances train = randData.trainCV(folds, n, rand);

        // build and evaluate classifier
        Classifier clsCopy = Classifier.makeCopy(cls);
        //System.out.print("Building classifier ");
        clsCopy.buildClassifier(train);
        if (clsCopy.toString().trim().equals("")) 
        	writer.write("Warning, evaluator has an empty regex\n");
        writer.write(clsCopy.toString() + "\n");
        //System.out.print("Evaluating classifier\n");
        eval.evaluateModel(clsCopy, test);
      }

      // output evaluation    
      writer.write("\n=== Setup run " + (i+1) + " ===\n");
      writer.write("\nSeed: " + seed);
      writer.write("\n");
      writer.write(eval.toSummaryString("\n=== " + folds + "-fold Cross-validation run " + (i+1) + "===", false));
      writer.write("\n" + eval.toMatrixString() + "\n");
    }
    writer.close();
  }
}
